"""Common test fixtures for the talus package."""
import json
from typing import Type
from uuid import uuid4

import pika
import pytest
from pydantic import Field

from talus.consumer import DurableConsumer
from talus.models.binding import Binding
from talus.models.connection_parameters import ConsumerConnectionParameterFactory
from talus.models.connection_parameters import ProducerConnectionParameterFactory
from talus.models.exchange import Exchange
from talus.models.message import ConsumeMessageBase
from talus.models.message import MessageBodyBase
from talus.models.message import PublishMessageBase
from talus.models.queue import Queue
from talus.models.retryer import ConnectionRetryerFactory
from talus.producer import DurableProducer


@pytest.fixture()
def message_data_schema_cls() -> Type[MessageBodyBase]:
    class MessageDataBody(MessageBodyBase):
        key: str = Field(default="value")

    return MessageDataBody


@pytest.fixture(params=["no-conversation-id", "conversation-id"])
def message_data(request, message_data_schema_cls) -> dict[str, str]:
    result = message_data_schema_cls(key=uuid4().hex).model_dump()
    match request.param:
        case "no-conversation-id":
            result.pop("conversationId")
            return result
        case "conversation-id":
            return result
        case _:
            raise NotImplementedError("message data parameter not implemented")


@pytest.fixture(params=["bytes-body", "python-body", "json-body"])
def body(message_data, request) -> bytes:
    match request.param:
        case "bytes-body":
            return json.dumps(message_data).encode("utf-8")
        case "python-body":
            return message_data
        case "json-body":
            return json.dumps(message_data)
        case _:
            raise NotImplementedError("body type not implemented")


@pytest.fixture(params=["no-headers", "headers"])
def headers(request) -> dict[str, str] | None:
    match request.param:
        case "no-headers":
            return None
        case "headers":
            return {"header1": "value1"}
        case _:
            raise NotImplementedError("headers parameter not implemented")


@pytest.fixture()
def properties(headers) -> pika.spec.BasicProperties:
    return pika.spec.BasicProperties(headers=headers)


@pytest.fixture()
def routing_key() -> str:
    return uuid4().hex


@pytest.fixture()
def delivery_tag() -> str:
    return uuid4().hex


@pytest.fixture()
def method(routing_key, delivery_tag) -> pika.spec.Basic.Deliver:
    return pika.spec.Basic.Deliver(routing_key=routing_key, delivery_tag=delivery_tag)


@pytest.fixture
def test_queue(request) -> Queue:
    """A queue name for consume tests."""
    prefix = request.node.name
    return Queue(name=f"queue_{prefix}_{uuid4().hex[:6]}", auto_delete=True)


@pytest.fixture
def consumer(test_queue) -> DurableConsumer:
    """A test consumer."""
    with DurableConsumer(
        consume_queue=test_queue,
        prefetch_count=1,
        connection_parameters=ConsumerConnectionParameterFactory(),
        connection_retryer=ConnectionRetryerFactory(attempts=2, wait=0),
    ) as consumer:
        yield consumer


@pytest.fixture()
def consume_message_cls(message_data_schema_cls) -> Type[ConsumeMessageBase]:
    """A test consume message class."""

    class ConsumeMessage(ConsumeMessageBase):
        message_body_cls = message_data_schema_cls

    return ConsumeMessage


@pytest.fixture
def publish_message_cls(routing_key, headers) -> Type[PublishMessageBase]:
    """A test publish message class."""
    rk = routing_key
    h = headers

    class PublishMessage(PublishMessageBase):
        default_routing_key: str = rk
        headers: dict[str, str] = h

    return PublishMessage


@pytest.fixture()
def direct_exchange() -> Exchange:
    return Exchange(
        name=f"direct_{uuid4().hex[:6]}",
        auto_delete=True,
    )


@pytest.fixture()
def queue_bindings(test_queue, publish_message_cls) -> Binding:
    """A list of routing key -> queue bindings"""
    return Binding(
        message=publish_message_cls,
        queue=test_queue,
    )


@pytest.fixture
def producer(queue_bindings, direct_exchange) -> DurableProducer:
    """A test producer."""

    with DurableProducer(
        queue_bindings=queue_bindings,
        publish_exchange=direct_exchange,
        connection_parameters=ProducerConnectionParameterFactory(),
        connection_retryer=ConnectionRetryerFactory(attempts=2, wait=0),
    ) as producer:
        yield producer


@pytest.fixture
def single_message_on_queue(producer, test_queue, publish_message_cls, body) -> PublishMessageBase:
    """
    Fixture that produces a single message to an empty test queue
    """
    publish_message = publish_message_cls(body=body)
    producer.publish(publish_message)
    return publish_message



"""Base Processor for messages consumed from a channel."""
import logging
from abc import ABC
from abc import abstractmethod
from typing import Type

import pika
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import ValidationError

from talus.models.message import ConsumeMessageBase
from talus.producer import DurableProducer

logger = logging.getLogger(__name__)


DEFAULT_DEAD_LETTER_EXCEPTIONS: tuple[type[Exception]] = (ValidationError,)


def format_dead_letter_exceptions(
    exceptions: Type[Exception] | tuple[Type[Exception]] | None = None,
) -> tuple[Type[Exception], ...]:
    """Append default exceptions to the given exceptions."""
    if exceptions is None:
        return DEFAULT_DEAD_LETTER_EXCEPTIONS
    if not isinstance(exceptions, tuple):
        exceptions = (exceptions,)
    return exceptions + DEFAULT_DEAD_LETTER_EXCEPTIONS


class MessageProcessorBase(BaseModel, ABC):
    """
    Base Processor for processing messages consumed from a channel.  This is pydantic.BaseModel to enhance
    validation and error handling.
    Implementations should extend this class and implement the process_message method.
    Instances of a processor will be used to process multiple messages when they are run.
    The signature of the of running a message processor instance matches the
    DurableConsumer().consume() method callback.
    :param: message_cls: Validate the message body with this class.
    :param: dead_letter_exceptions: Exceptions that will cause the message to be dead lettered.
    :param: producer: Producer to use for publishing messages as part of inbound message processing if necessary.
    >>> class ObjectBucketBody(MessageBodyBase):
    >>>    objectName: str
    >>>    bucket: str
    >>>
    >>> class InboundMessage(ConsumeMessageBase):
    >>>     message_body_cls: Type[ObjectBucketBody] = ObjectBucketBody  # Validate the body with ObjectBucketBody
    >>>
    >>> class MyMessageProcessor(MessageProcessorBase):
    >>>     message_cls: Type[InboundMessage] = InboundMessage
    >>>     def process_message(self, message: InboundMessage):
    >>>         print(f"Processing message: {message}")
    >>>
    >>> queue = Queue(name="my_queue")
    >>> with DurableConsumer(consume_queue=queue) as consumer:
    >>>     consumer.listen(MyMessageProcessor())
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    message_cls: Type[ConsumeMessageBase] = ConsumeMessageBase
    dead_letter_exceptions: tuple[Type[Exception]] = format_dead_letter_exceptions()
    producer: DurableProducer | None = None

    @abstractmethod
    def process_message(self, message: ConsumeMessageBase):
        pass  # pragma: no cover

    @classmethod
    def dlq_message(
        cls,
        channel: pika.adapters.blocking_connection.BlockingChannel,
        method: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        exception: Exception | None = None,
    ) -> None:
        """
        Reject the message but do not requeue it. Depending on exchange policy, this will trigger dead lettering
        https://www.rabbitmq.com/docs/dlx
        """
        logger.warning(f"Dead lettering message:{method=}, {properties=}, {exception=}")
        channel.basic_nack(delivery_tag=method.delivery_tag, requeue=False)

    @classmethod
    def acknowledge_message(
        cls, channel: pika.adapters.blocking_connection.BlockingChannel, message: ConsumeMessageBase
    ) -> None:
        """Acknowledge the message which removes it from the queue."""
        logger.debug(
            f"Acknowledging message: method={message.method}, properties={message.properties}, {message=}"
        )
        channel.basic_ack(delivery_tag=message.delivery_tag)

    def __call__(
        self,
        channel: pika.adapters.blocking_connection.BlockingChannel,
        method: pika.spec.Basic.Deliver,
        properties: pika.spec.BasicProperties,
        body: bytes,
    ) -> "MessageProcessorBase":
        """
        Message processors are meant to be called from the DurableConsumer.consume() method.
        Calling a message processor instance first validates the message body using the instance
        message_cls.  If the message body is valid, the process_message method is called with the
        validated model, otherwise the error is evaluated for dead lettering.  The default is to
        dead letter messages that fail validation but other exceptions can be added to the instance.
        Exceptions that are raised but not dead lettered will be re-raised while dead lettered
        exceptions will be logged but effectively be consumed.
        """
        try:
            message = self.message_cls(method=method, properties=properties, body=body)
            self.process_message(message)
            self.acknowledge_message(channel=channel, message=message)
        except self.dead_letter_exceptions as e:
            self.dlq_message(channel=channel, method=method, properties=properties, exception=e)
        return self


class DLQMessageProcessor(MessageProcessorBase):
    """
    A message processor that dead letters all messages it processes.  This is useful for testing and
    fall through cases where a message processor is not found for a message.
    """

    message_cls: Type[ConsumeMessageBase] = ConsumeMessageBase
    dead_letter_exceptions: tuple[Type[Exception]] = format_dead_letter_exceptions(ValueError)

    def process_message(self, message: ConsumeMessageBase):
        raise ValueError(f"DLQMessageProcessor auto fails processing: {message=}")



"""Tests for the connection parameters factory"""
import pika

from talus.models.connection_parameters import ConnectionParameterFactory


def test_connection_parameters():
    """
    :given: A connection parameter factory
    :when: The factory is called
    :then: A pika.ConnectionParameters object is returned
    """
    connection_parameters = ConnectionParameterFactory()
    assert isinstance(connection_parameters(), pika.ConnectionParameters)



"""Top level api"""
from talus.consumer import DurableConsumer
from talus.models.binding import Binding
from talus.models.connection_parameters import ConsumerConnectionParameterFactory
from talus.models.connection_parameters import ProducerConnectionParameterFactory
from talus.models.exchange import Exchange
from talus.models.message import ConsumeMessageBase
from talus.models.message import MessageBodyBase
from talus.models.message import PublishMessageBase
from talus.models.processor import MessageProcessorBase
from talus.models.queue import Queue
from talus.models.retryer import ConnectionRetryerFactory
from talus.producer import DurableProducer



"""Test the durable connection base module."""
import logging

import pika
import pytest
from tenacity import after_log
from tenacity import retry_if_exception_type
from tenacity import Retrying
from tenacity import stop_after_attempt
from tenacity import wait_exponential
from tenacity import wait_random

from talus.base import DurableConnection
from talus.models.connection_parameters import ConnectionParameterFactory
from talus.models.retryer import ConnectionRetryerFactory
from talus.models.retryer import DEFAULT_CONNECTION_EXCEPTIONS


logger = logging.getLogger(__name__)


@pytest.fixture(params=["pika.ConnectionParameters", "ConnectionParameterFactory"])
def connection_parameters(request):
    match request.param:
        case "pika.ConnectionParameters":
            return pika.ConnectionParameters(
                host="localhost",
                port=5672,
                credentials=pika.PlainCredentials("guest", "guest"),
            )
        case "ConnectionParameterFactory":
            return ConnectionParameterFactory()
        case _:
            raise NotImplementedError("Fixture parameter not implemented.")


@pytest.fixture(params=["Retrying", "RetryerFactory"])
def connection_retryer(request):
    match request.param:
        case "Retrying":
            return Retrying(
                retry=retry_if_exception_type(DEFAULT_CONNECTION_EXCEPTIONS),
                wait=wait_exponential(
                    multiplier=1,
                    min=1,
                    max=3,
                )
                + wait_random(1, 3),
                stop=stop_after_attempt(2),
                after=after_log(logger=logger, log_level=logging.INFO),
            )
        case "RetryerFactory":
            return ConnectionRetryerFactory(
                delay_min=1,
                delay_max=3,
                jitter_min=1,
                jitter_max=3,
                attempts=2,
                exceptions=DEFAULT_CONNECTION_EXCEPTIONS,
            )
        case _:
            raise NotImplementedError("Fixture parameter not implemented.")


def test_durable_connection(connection_parameters, connection_retryer):
    """
    :given: Connection Parameters and Retryer
    :when: DurableConnection is instantiated
    :then: Connection can be established.
    """
    # when
    with DurableConnection(
        connection_parameters=connection_parameters, connection_retryer=connection_retryer
    ) as durable_connection:
        # then
        durable_connection.connect()
        assert durable_connection.is_connected
    assert not durable_connection.is_connected


@pytest.fixture()
def durable_connection(connection_parameters, connection_retryer):
    with DurableConnection(
        connection_parameters=connection_parameters, connection_retryer=connection_retryer
    ) as durable_connection:
        yield durable_connection


def test_create_queue(durable_connection, test_queue):
    """
    :given: Connection to RabbitMQ
    :when: Queue is created
    :then: Queue is created
    """
    # given
    durable_connection.connect()
    # when
    durable_connection.create_queue(test_queue)
    # then
    test_queue.passive = True  # Only check if queue exists and raise `ChannelClosed` if it doesn't
    durable_connection.create_queue(test_queue)


def test_repr(durable_connection):
    """
    :given: Connection to RabbitMQ
    :when: Connection is represented as a string
    :then: String representation is returned
    """
    # when
    assert DurableConnection.__name__ in repr(durable_connection)



"""
Base wrapper definition shared by both consumer and producer wrappers
"""
import logging
from typing import Callable

import pika
from tenacity import Retrying

from talus.models.connection_parameters import ConnectionParameterFactory
from talus.models.queue import Queue
from talus.models.retryer import ConnectionRetryerFactory

logger = logging.getLogger(__name__)


class DurableConnection:
    """
    RabbitMQ connector that establishes a blocking connection and channel
    """

    def __init__(
        self,
        connection_parameters: pika.ConnectionParameters
        | Callable[[], pika.ConnectionParameters] = ConnectionParameterFactory(),
        connection_retryer: Retrying | Callable[[], Retrying] = ConnectionRetryerFactory(),
    ):
        """
        Constructor for a durable connection to RabbitMQ.
        :param connection_parameters: A pika.ConnectionParameters object or one resulting
            from a callable that returns a ConnectionParameters object.
        :param connection_retryer: A tenacity.Retrying object or one resulting from a callable that returns
            a Retrying object.
        """
        if isinstance(connection_parameters, pika.ConnectionParameters):
            self.connection_parameters = connection_parameters
        else:  # factory
            self.connection_parameters = connection_parameters()
        if isinstance(connection_retryer, Retrying):
            self.connection_retryer = connection_retryer
        else:  # factory
            self.connection_retryer = connection_retryer()
        self.connection: pika.BlockingConnection | None = None
        self.channel: pika.adapters.blocking_connection.BlockingChannel | None = None

    def connect(self):
        """
        Connect to the RabbitMQ server retrying errors configured in the connection_retryer.
        """
        if not self.is_connected:
            self.connection_retryer(self._connect)

    def _connect(self):
        """
        Configures and initiates connection to the RabbitMQ server.
        """
        logger.debug(
            f"Attempt to connect to RabbitMQ: connection_params={self.connection_parameters}"
        )
        self.connection = pika.BlockingConnection(self.connection_parameters)
        logger.info(f"Connection Created")
        self.channel = self.connection.channel()
        logger.info("Channel Created")
        logger.info(f"Connected to RabbitMQ: connection={self.connection_parameters}")

    @property
    def is_connected(self):
        """
        Current state of the connection.  Only updated when the connection is used.

        :return: Latest connection state
        """
        if self.connection is not None:
            return self.connection.is_open
        return False

    def disconnect(self):
        """
        Closes connection and related channels to the RabbitMQ Server.
        """

        if self.is_connected:
            self.connection.close()
        logger.info(f"Disconnected from RabbitMQ: " f"connection={self.connection_parameters}")

    def create_queue(self, queue: Queue):
        """
        Create a queue on an already opened connection.
        """
        self.channel.queue_declare(
            queue=queue.name,
            durable=queue.durable,
            passive=queue.passive,
            auto_delete=queue.auto_delete,
            exclusive=queue.exclusive,
            arguments=queue.arguments,
        )
        logger.info(f"Queue Created: queue={queue.name}")

    def __enter__(self):
        """
        Entry for context manager.

        :return: connected instance of self
        """
        self.connect()
        return self

    def __exit__(self, exc_type, value, traceback):
        """
        Exit for context manager which disconnects from rabbitmq
        """
        self.disconnect()

    def __repr__(self):
        """Representation of the DurableConnection object"""
        return f"{self.__class__.__name__}(connection_parameters={self.connection_parameters!r}), connection_retryer={self.connection_retryer!r})"



"""Tests for the various models in the Talus application."""
from uuid import uuid4

import pytest

from talus.models.binding import Binding
from talus.models.message import ConsumeMessageBase
from talus.models.message import PublishMessageBase
from talus.models.queue import Queue


@pytest.fixture()
def queue_name() -> str:
    return uuid4().hex


@pytest.fixture()
def queue(queue_name) -> Queue:
    return Queue(name=queue_name)


def test_binding_valid(queue, queue_name, publish_message_cls, routing_key):
    """
    :given: A message class and a queue.
    :when: A binding is created.
    :then: the routing key and queue name are accessible.
    """
    # when
    binding = Binding(message=publish_message_cls, queue=queue)
    # then
    assert binding.routing_key == routing_key
    assert binding.queue_name == queue_name


@pytest.fixture()
def consume_message(body, method, properties) -> ConsumeMessageBase:
    return ConsumeMessageBase(method=method, properties=properties, body=body)


class BadPublishMessage(PublishMessageBase):
    default_routing_key = ""


@pytest.mark.parametrize(
    "message, queue",
    [
        pytest.param(
            ConsumeMessageBase,
            Queue(name="foo"),
            id="consume_message",
        ),
        pytest.param(
            BadPublishMessage,
            Queue(name="foo"),
            id="routing_key_empty",
        ),
    ],
)
def test_binding_invalid(message, queue):
    """
    :given: A message without a routing key and a queue.
    :when: A binding is created.
    :then: A Value error is raised
    """
    # when/then
    with pytest.raises(ValueError):
        Binding(message=message, queue=queue)



"""Tests for the message processor models."""
from typing import Type

import pika
import pytest

from talus.models.message import ConsumeMessageBase
from talus.models.processor import DEFAULT_DEAD_LETTER_EXCEPTIONS
from talus.models.processor import DLQMessageProcessor
from talus.models.processor import format_dead_letter_exceptions
from talus.models.processor import MessageProcessorBase


@pytest.fixture()
def successful_message_processor(consume_message_cls) -> MessageProcessorBase:
    """
    A message processor that does not raise an exception.
    """

    class SuccessfulMessageProcessor(MessageProcessorBase):
        message_cls: Type[ConsumeMessageBase] = consume_message_cls

        def process_message(self, message: ConsumeMessageBase):
            self._message_was_processed = True

    yield SuccessfulMessageProcessor()


@pytest.fixture()
def consumed_message(
    publish_message_cls,
    producer,
    consumer,
    body,
) -> tuple[pika.spec.Basic.Deliver, pika.spec.BasicProperties, bytes]:
    outbound_message = publish_message_cls(body=body)
    producer.publish(message=outbound_message)
    for m, p, b in consumer.consume_generator(auto_ack=True):
        yield m, p, b  # only the first one
        break


def test_message_processor_success(consumer, consumed_message, successful_message_processor):
    """
    :given: A message processor that does not raise an exception.
    :when: The message processor is called.
    :then: The message processor should process and acknowledge the message.
    """
    # given
    method, properties, body = consumed_message
    # when
    successful_message_processor(
        channel=consumer.channel, method=method, properties=properties, body=body
    )
    # then
    assert successful_message_processor._message_was_processed
    assert not consumer.channel.get_waiting_message_count()  # no messages left in the queue


@pytest.fixture()
def erroring_message_processor(consume_message_cls) -> MessageProcessorBase:
    """
    A message processor that raises a non-dead-letter exception.
    """

    class ErroringMessageProcessor(MessageProcessorBase):
        message_cls: Type[ConsumeMessageBase] = consume_message_cls
        dead_letter_exceptions: Type[Exception] = ValueError

        def process_message(self, message: ConsumeMessageBase):
            raise RuntimeError("A bad message processor raised an exception that is not DLQ'd.")

    yield ErroringMessageProcessor()


def test_message_processor_failure(erroring_message_processor, consumed_message, consumer):
    """
    :given: A message processor that raises an exception.
    :when: The message processor is called.
    :then: Expected exception is raised.
    """
    # given
    method, properties, body = consumed_message
    # when/then
    with pytest.raises(RuntimeError):
        erroring_message_processor(
            channel=consumer.channel, method=method, properties=properties, body=body
        )


@pytest.fixture()
def dlq_message_processor(consume_message_cls) -> DLQMessageProcessor:
    """
    A message processor that raises a non-dead-letter exception.
    """

    class AutoDLQMessageProcessor(DLQMessageProcessor):
        message_cls: Type[ConsumeMessageBase] = consume_message_cls

        def process_message(self, message: ConsumeMessageBase):
            self._process_message_called = True
            super().process_message(message)

    yield AutoDLQMessageProcessor()


def test_dlq_message_processor(dlq_message_processor, consumed_message, consumer):
    """
    :given: A message processor that raises an exception identified as for DLQ.
    :when: The message processor is called.
    :then: The message processor should DLQ the message.
    """
    # given
    method, properties, body = consumed_message
    # when
    dlq_message_processor(channel=consumer.channel, method=method, properties=properties, body=body)
    # then
    assert dlq_message_processor._process_message_called
    assert not consumer.channel.get_waiting_message_count()  # no messages left in the queue


@pytest.fixture()
def message_processor_with_producer(
    producer, queue_bindings, direct_exchange
) -> MessageProcessorBase:
    """
    A message processor that has a producer.
    """
    durable_producer = producer

    class MessageProcessorWithProducer(MessageProcessorBase):
        message_cls: Type[ConsumeMessageBase] = ConsumeMessageBase

        def process_message(self, message: ConsumeMessageBase):
            if self.producer.is_connected:
                self._message_was_processed = True

    return MessageProcessorWithProducer(producer=durable_producer)


def test_message_processor_with_producer_success(
    consumer, consumed_message, message_processor_with_producer
):
    """
    :given: A message processor that does not raise an exception.
    :when: The message processor is called.
    :then: The message processor should process and acknowledge the message.
    """
    # given
    method, properties, body = consumed_message
    # when
    message_processor_with_producer(
        channel=consumer.channel, method=method, properties=properties, body=body
    )
    # then
    assert message_processor_with_producer._message_was_processed
    assert not consumer.channel.get_waiting_message_count()  # no messages left on queue


@pytest.mark.parametrize(
    "exc",
    [
        pytest.param(KeyError, id="single"),
        pytest.param(None, id="none"),
        pytest.param((ValueError, RuntimeError), id="multiple"),
    ],
)
def test_format_dead_letter_exceptions(exc):
    """
    Given: an exception
    When: the exception is formatted
    Then: the default exceptions are added to the formatted exception
    """
    result = format_dead_letter_exceptions(exc)
    assert isinstance(result, tuple)
    for e in DEFAULT_DEAD_LETTER_EXCEPTIONS:
        assert e in result
    if exc is not None:
        if isinstance(exc, tuple):
            for e in exc:
                assert e in result
        else:
            assert exc in result



"""Connection Parameter Factory Model"""
import pika
from pydantic import BaseModel
from pydantic import Field


MINUTES_1 = 60
HOUR_1 = 3600


class ConnectionParameterFactory(BaseModel):
    """
    Translator for the connection configuration to a pika.ConnectionParameters object.
    >>> from talus.models.connection_parameters import ConnectionParameterFactory
    >>> factory = ConnectionParameterFactory()
    >>> connection_parameters = factory()
    """

    rabbitmq_host: str = Field(
        default="127.0.0.1"
    )  # Host name or IP of the rabbitMQ server. e.g. 127.0.0.1
    rabbitmq_port: int = 5672  # Port the rabbitmq server listens on e.g. 5672
    rabbitmq_user: str = Field(default="guest")  # Username for the rabbitMQ server e.g. guest
    rabbitmq_pass: str = Field(default="guest")  # Password for the rabbitMQ server e.g. guest
    connection_name: str = Field(
        default="default_connection"
    )  # Name of the connection that will be visible in the rabbitmq admin console
    # Controls AMQP heartbeat timeout negotiation
    # during connection tuning. An integer value always overrides the value
    # proposed by broker. Use 0 to deactivate heartbeats and None to always
    # accept the broker's proposal.
    heartbeat: int | None = None

    def __call__(self) -> pika.ConnectionParameters:
        return pika.ConnectionParameters(
            host=self.rabbitmq_host,
            port=self.rabbitmq_port,
            credentials=pika.credentials.PlainCredentials(self.rabbitmq_user, self.rabbitmq_pass),
            heartbeat=self.heartbeat,
            client_properties={"connection_name": self.connection_name},
        )


class ConsumerConnectionParameterFactory(ConnectionParameterFactory):
    """Consumer Connection Parameter Factory Model which defaults to a 1 minute heartbeat."""

    heartbeat: int = MINUTES_1


class ProducerConnectionParameterFactory(ConnectionParameterFactory):
    """Producer Connection Parameter Factory Model which defaults to a 1 hour heartbeat."""

    heartbeat: int = HOUR_1



"""
Producer implementation of the base connection wrapper
"""
import logging
from typing import Callable

import pika
from pika.exceptions import NackError
from pika.exceptions import UnroutableError
from tenacity import Retrying

from talus.base import DurableConnection
from talus.models.binding import Binding
from talus.models.connection_parameters import ProducerConnectionParameterFactory
from talus.models.exchange import Exchange
from talus.models.message import PublishMessageBase
from talus.models.retryer import ConnectionRetryerFactory
from talus.models.retryer import RetryerFactory

logger = logging.getLogger(__name__)


class DurableProducer(DurableConnection):
    """
    RabbitMQ Connector for posting to 1 to many queues via an exchange
    """

    def __init__(
        self,
        queue_bindings: list[Binding] | Binding,
        publish_exchange: Exchange,
        connection_parameters: pika.ConnectionParameters
        | Callable[[], pika.ConnectionParameters] = ProducerConnectionParameterFactory(),
        connection_retryer: Retrying | Callable[[], Retrying] = ConnectionRetryerFactory(),
    ):
        """
        Constructor for the producer connector

        :param queue_bindings: Bindings used to map routing key to destination queue for an exchange

        :param publish_exchange: Name of the exchange that the  producer will publish to.

        :param connection_parameters: A pika.ConnectionParameters object or one resulting
            from a callable that returns a ConnectionParameters object.

        :param connection_retryer: A tenacity.Retrying object or one resulting from a callable that returns
            a Retrying object.

        >>> from talus.models.binding import Binding
        >>> from talus.models.exchange import Exchange
        >>> from talus.producer import DurableProducer
        >>> from talus.models.retryer import ConnectionRetryerFactory
        >>> from talus.models.connection_parameters import ProducerConnectionParameterFactory
        >>> from talus.models.queue import Queue
        >>> from talus.models.message import PublishMessageBase
        >>> queue = Queue(name="test_queue")
        >>> exchange = Exchange(name="test_exchange")
        >>> # Bind the queue to the routing key on message.
        >>> queue_bindings = Binding(queue=queue, message=PublishMessageBase)
        >>> with DurableProducer(queue_bindings=queue_bindings, publish_exchange=exchange, connection_parameters=ProducerConnectionParameterFactory(), connection_retryer=ConnectionRetryerFactory()) as producer:
        >>>     producer.publish(PublishMessageBase(body={"test": "test"}))

        """
        super().__init__(
            connection_parameters=connection_parameters, connection_retryer=connection_retryer
        )
        if isinstance(queue_bindings, Binding):
            queue_bindings = [queue_bindings]
        self.queue_bindings = queue_bindings
        self.publish_exchange = publish_exchange

        # Retry posts which are not confirmed by the broker
        publish_retryer_factory = RetryerFactory(
            delay_min=1,
            delay_max=3,
            jitter_min=1,
            jitter_max=3,
            attempts=3,
            exceptions=(UnroutableError, NackError),
        )
        self.publish_retryer = publish_retryer_factory()

    def _connect(self) -> None:
        """
        Configures and initiates producer connection to the RabbitMQ server.
        :return:
        """
        super()._connect()
        self.channel.confirm_delivery()  # ensure persistence prior to message confirmation
        self.channel.exchange_declare(
            exchange=self.publish_exchange.name,
            exchange_type=self.publish_exchange.type,
            passive=self.publish_exchange.passive,
            durable=self.publish_exchange.durable,
            auto_delete=self.publish_exchange.auto_delete,
            internal=self.publish_exchange.internal,
            arguments=self.publish_exchange.arguments,
        )
        for binding in self.queue_bindings:
            self.create_queue(queue=binding.queue)
            self.channel.queue_bind(
                exchange=self.publish_exchange.name,
                queue=binding.queue_name,
                routing_key=binding.routing_key,
            )
            logger.info(
                f"Bindings configured: exchange={self.publish_exchange}, "
                f"queue={binding.queue.name}, "
                f"routing_key={binding.routing_key} "
            )

    def _publish(self, message: PublishMessageBase) -> None:
        """
        Publish message to the exchange configured on the producer

        :param message: Message to publish
        :return: None
        """
        self.connect()
        self.publish_retryer(
            self.channel.basic_publish,
            exchange=self.publish_exchange.name,
            routing_key=message.routing_key,
            body=message.body.model_dump_json(by_alias=True),
            properties=message.properties,
            mandatory=True,  # paired with channel.confirm_delivery() to ensure message persistence
        )

    def publish(self, message: PublishMessageBase) -> None:
        """
        Publish message to the exchange configured on the producer retrying connection errors configured in the connection_retryer.
        :return: None
        """
        self.connection_retryer(self._publish, message=message)

    def __repr__(self):
        """Representation of the DurableProducer instance."""
        return f"{self.__class__.__name__}(queue_bindings={self.queue_bindings !r}, exchange={self.publish_exchange!r}, connection_parameters={self.connection_parameters!r}, connection_retryer={self.connection_retryer!r})"
